_TRACKING_METHODS_REGISTRY = {}


class BaseTrackingMethod:
  def __init__(self, videoPath: str, wellPositions: list, hyperparameters: dict):
    '''
    Initialize tracking method.

    Parameters:
    videoPath - full path to the video
    wellPositions - a list of dicts containing keys TopLeftX, TopLeftY, LengthX, LengthY, representing the wells selected by the user
    hyperparameters - a dict containing parameter values based on the config file
    '''
    raise NotImplementedError

  def run(self) -> dict:
    '''
    Run tracking and return output in the designated format.

    Expected output format:
     - a dict mapping well numbers to well data.
     - well data is a list of dicts, each containing information about a single bout of movement
     - bout dict must contain the following keys:
       - AnimalNumber: an integer
       - BoutStart: an integer, the frame in which the bout started
       - BoutEnd: an integer, the frame in which the bout ended
       - HeadX: a list of numbers representing the X coordinates of the center of the head, one for each frame of the bout
       - HeadY: a list of numbers representing the Y coordinates of the center of the head, one for each frame of the bout       
    - depending on the type of tracking performed, it can also contain:
       - Heading: a list of main head axis angles, one for each frame of the bout
       - TailAngleRaw: a list of tail angles (see figure 1c in the original 2013 ZebraZoom paper), one for each frame of the bout (if tail tracking is performed)
       - TailX_VideoReferential: a list in which each element corresponds to a frame, with each of those element itself being a list containing all x axis coordinates of points along the tail (if tail tracking is performed)
       - TailY_VideoReferential: a list in which each element corresponds to a frame, with each of those element itself being a list containing all y axis coordinates of points along the tail (if tail tracking is performed)
    - if eyes tracking if performed:
       - leftEyeX, leftEyeY, leftEyeAngle, leftEyeArea, rightEyeX, rightEyeY, rightEyeAngle, rightEyeArea: lists of integers corresponding to the x and y coordinates, main angle and eye area of the left and right eye for each frame of the bout
    '''
    raise NotImplementedError


def register_tracking_method(name: str, factory: BaseTrackingMethod) -> None:
  '''Register a new tracking method.'''
  _TRACKING_METHODS_REGISTRY[name] = factory


def get_default_tracking_method() -> BaseTrackingMethod:
  return _TRACKING_METHODS_REGISTRY['tracking']


def get_tracking_method(method: str) -> BaseTrackingMethod:
  return _TRACKING_METHODS_REGISTRY[method]



import numpy as np
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.preprocessImage import preprocessImage
import math
import json
import random
import os


def improveContrast(frame, hyperparameters):
  frame = 255 - frame
  frameIsBGR = len(frame.shape) == 3
  if frameIsBGR:
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
  lowVal = int(np.quantile(frame, hyperparameters["outputValidationVideoContrastImprovementQuartile"]))
  highVal = int(np.quantile(frame, 1 - hyperparameters["outputValidationVideoContrastImprovementQuartile"]))
  frame[frame < lowVal] = lowVal
  frame[frame > highVal] = highVal
  frame = frame - lowVal
  mult = np.max(frame)
  frame = frame * (255 / mult)
  frame = frame.astype('uint8')
  if frameIsBGR:
    frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)
  return frame


def calculateInfoFrameForFrame(superStruct, hyperparameters, wellIdx, animalIdx, boutIdx, frameIdx, colorModifTab):
  plotOnlyOneTailPointForVisu = hyperparameters["plotOnlyOneTailPointForVisu"]
  trackingPointSizeDisplay    = hyperparameters["trackingPointSizeDisplay"]
  infoWells = []
  
  for i in range(0, len(superStruct["wellPositions"])):
    x = superStruct["wellPositions"][i]["topLeftX"]
    y = superStruct["wellPositions"][i]["topLeftY"]
    infoWells.append([x, y])

  infoFrame = []
  if hyperparameters['extractAdvanceZebraParameters'] and ("Bend_TimingAbsolute" in superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]):
    Bend_TimingAbsolute = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["Bend_TimingAbsolute"]
  else:
    Bend_TimingAbsolute = []

  relativeFrameIdx = frameIdx - superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["BoutStart"]

  anglePointsBefore = []
  anglePointsBaseBefore = []
  x0 = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["HeadX"][0]
  y0 = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["HeadY"][0]

  x = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["HeadX"][relativeFrameIdx]
  y = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["HeadY"][relativeFrameIdx]
  Heading = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["Heading"][relativeFrameIdx] if "Heading" in superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx] else -10

  x = x + infoWells[wellIdx][0]
  y = y + infoWells[wellIdx][1]
  dataToPlot = {}
  dataToPlot["x"]       = x
  dataToPlot["y"]       = y
  dataToPlot["size"]    = trackingPointSizeDisplay + (0 if plotOnlyOneTailPointForVisu else 2)
  dataToPlot["Heading"] = Heading
  dataToPlot["numMouv"] = boutIdx+1
  dataToPlot["numWell"] = wellIdx
  dataToPlot["numAnimal"] = animalIdx

  dataToPlot["red"]     = 0
  dataToPlot["green"]   = 100
  dataToPlot["blue"]    = 255

  infoFrame.append(dataToPlot)

  # Points along the Tail
  nbPointsToPlot   = 0
  if plotOnlyOneTailPointForVisu:
    nbPointsToPlot = len(superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["TailX_VideoReferential"][relativeFrameIdx])
    pointsToPlot = (0, *range(nbPointsToPlot - 1, nbPointsToPlot))
  else:
    nbPointsToPlot = len(superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["TailX_VideoReferential"][relativeFrameIdx]) if "TailX_VideoReferential" in superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx] else 0
    pointsToPlot = range(nbPointsToPlot)

  for m in pointsToPlot:
    tailX = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["TailX_VideoReferential"][relativeFrameIdx][m]
    tailY = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["TailY_VideoReferential"][relativeFrameIdx][m]
    tailX = tailX + infoWells[wellIdx][0]
    tailY = tailY + infoWells[wellIdx][1]
    size = 1
    dataToPlot      = {}
    dataToPlot["x"] = tailX
    dataToPlot["y"] = tailY
    if (m == (nbPointsToPlot-1)):
      if (Bend_TimingAbsolute.count(frameIdx+1) != 0):
        dataToPlot["size"] = trackingPointSizeDisplay + (0 if plotOnlyOneTailPointForVisu else 5)
        dataToPlot["red"]   = 0
        dataToPlot["green"] = 0
        dataToPlot["blue"]  = 255
      else:
        dataToPlot["size"] = trackingPointSizeDisplay + (0 if plotOnlyOneTailPointForVisu else 2)
        dataToPlot["red"]   = 0
        dataToPlot["green"] = 255
        dataToPlot["blue"]  = 0
    else:
      dataToPlot["size"] = trackingPointSizeDisplay
      dataToPlot["red"]   = 0   + colorModifTab[animalIdx]["red"]
      dataToPlot["green"] = 255 - colorModifTab[animalIdx]["green"]
      dataToPlot["blue"]  = 0   + colorModifTab[animalIdx]["blue"]
    dataToPlot["Heading"] = -10

    infoFrame.append(dataToPlot)

  if hyperparameters["eyeTracking"] or "leftEyeX" in superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]:
    leftEyeX     = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["leftEyeX"][relativeFrameIdx]
    leftEyeY     = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["leftEyeY"][relativeFrameIdx]
    leftEyeAngle = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["leftEyeAngle"][relativeFrameIdx]
    leftEyeX = leftEyeX + infoWells[wellIdx][0]
    leftEyeY = leftEyeY + infoWells[wellIdx][1]
    dataToPlot = {}
    dataToPlot["x"]       = leftEyeX
    dataToPlot["y"]       = leftEyeY
    dataToPlot["size"]    = 1
    dataToPlot["Heading"] = leftEyeAngle
    dataToPlot["headingColor"] = (0, 255, 0)
    dataToPlot["headingWidth"]    = 2 # hyperparameters["eyeTrackingHeadEmbeddedWidth"]
    dataToPlot["headingHalfDiam"] = hyperparameters["eyeTrackingHeadEmbeddedHalfDiameter"]
    dataToPlot["numMouv"] = boutIdx+1
    dataToPlot["numWell"] = wellIdx
    dataToPlot["red"]     = 255
    dataToPlot["green"]   = 255
    dataToPlot["blue"]    = 0
    infoFrame.append(dataToPlot)

    rightEyeX     = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["rightEyeX"][relativeFrameIdx]
    rightEyeY     = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["rightEyeY"][relativeFrameIdx]
    rightEyeAngle = superStruct["wellPoissMouv"][wellIdx][animalIdx][boutIdx]["rightEyeAngle"][relativeFrameIdx]
    rightEyeX = rightEyeX + infoWells[wellIdx][0]
    rightEyeY = rightEyeY + infoWells[wellIdx][1]
    dataToPlot = {}
    dataToPlot["x"]       = rightEyeX
    dataToPlot["y"]       = rightEyeY
    dataToPlot["size"]    = 1
    dataToPlot["Heading"] = rightEyeAngle
    dataToPlot["headingColor"] = (0, 255, 0)
    dataToPlot["headingWidth"]    = 2 # hyperparameters["eyeTrackingHeadEmbeddedWidth"]
    dataToPlot["headingHalfDiam"] = hyperparameters["eyeTrackingHeadEmbeddedHalfDiameter"]
    # dataToPlot["numMouv"] = boutIdx+1
    # dataToPlot["numWell"] = wellIdx
    dataToPlot["red"]     = 0
    dataToPlot["green"]   = 0
    dataToPlot["blue"]    = 255
    infoFrame.append(dataToPlot)
  return infoFrame

def calculateInfoFrame(superStruct, hyperparameters, nbFrames, colorModifTab):
  infoFrame = [[] for _ in range(nbFrames)]

  for i in range(0, len(superStruct["wellPoissMouv"])):
    for j in range(0, len(superStruct["wellPoissMouv"][i])):
      for k in range(0, len(superStruct["wellPoissMouv"][i][j])):
        BoutStart = superStruct["wellPoissMouv"][i][j][k]["BoutStart"]
        BoutEnd = superStruct["wellPoissMouv"][i][j][k]["BoutEnd"]
        mouvLength = len(superStruct["wellPoissMouv"][i][j][k]["HeadX"])
        for l in range(BoutStart, min(BoutEnd + 1, nbFrames)):
          infoFrame[l].extend(calculateInfoFrameForFrame(superStruct, hyperparameters, i, j, k, l, colorModifTab))
  return infoFrame


def drawInfoFrame(frame, infoFrame, colorModifTab, hyperparameters):
  for data in infoFrame:
    x = data["x"]
    y = data["y"]
    size  = data["size"]
    red   = data["red"]
    green = data["green"]
    blue  = data["blue"]
    if (data["Heading"] != -10):
      heading = data["Heading"]
      headingColor = data["headingColor"] if "headingColor" in data else (255,0,0)
      headingWidth = data["headingWidth"] if "headingWidth" in data else hyperparameters["trackingPointSizeDisplay"]
      headingHalfDiam = data["headingHalfDiam"] if "headingHalfDiam" in data else 20
      if hyperparameters["validationVideoPlotHeading"] and not(np.isnan(x)) and not(np.isnan(y)) and not(np.isnan(heading)):
        if hyperparameters["debugValidationVideoHeading"] == 0:
          cv2.line(frame,(int(x),int(y)),(int(x+headingHalfDiam*math.cos(heading)),int(y+headingHalfDiam*math.sin(heading))), headingColor, headingWidth)
        else:
          cv2.line(frame,(int(x),int(y)),(int(x-250*math.cos(heading)),int(y-250*math.sin(heading))), headingColor, headingWidth)

      # if ("numMouv" in data) and ("numWell" in data):
        # numMouv = data["numMouv"]
        # numWell = data["numWell"]
        # cv2.putText(frame,str(numMouv),(15+infoWells[numWell][0],25+infoWells[numWell][1]),cv2.FONT_HERSHEY_SIMPLEX,1,(255,255,255))
    
    if x != float('nan') and y != float('nan') and not(math.isnan(x)) and not(math.isnan(y)):
      cv2.circle(frame, (int(x),int(y)), size, (red,green,blue), -1)

    if hyperparameters["validationVideoPlotAnimalNumber"]:
      if "numAnimal" in data:
        numAnimal = int(data["numAnimal"])
        red       = int(0   + colorModifTab[numAnimal]["red"])
        green     = int(255 - colorModifTab[numAnimal]["green"])
        blue      = int(0   + colorModifTab[numAnimal]["blue"])
        cv2.putText(frame, str(numAnimal), (int(x + 10), int(y + 10)), cv2.FONT_HERSHEY_SIMPLEX, 1, (red, green, blue), 2)


def processFrame(frame, hyperparameters, infoFrame, colorModifTab):
  if hyperparameters["imagePreProcessMethod"]:
    frame = preprocessImage(frame, hyperparameters)

  if hyperparameters["reduceImageResolutionPercentage"]:
    frame_height, frame_width = frame.shape[:2]
    frame = cv2.resize(frame, (int(frame_width * hyperparameters["reduceImageResolutionPercentage"]), int(frame_height * hyperparameters["reduceImageResolutionPercentage"])), interpolation = cv2.INTER_AREA)
  if hyperparameters["outputValidationVideoContrastImprovement"]:
    frame = improveContrast(frame, hyperparameters)

  drawInfoFrame(frame, infoFrame, colorModifTab, hyperparameters)
  return frame


def createValidationVideo(videoPath, superStruct, hyperparameters, outputName=None):

  if (hyperparameters["freqAlgoPosFollow"] != 0):
    print("Starting the creation of the validation video")

  firstFrame                  = hyperparameters["firstFrame"]
  lastFrame                   = hyperparameters["lastFrame"]

  cap = zzVideoReading.VideoCapture(videoPath, hyperparameters)
  if (cap.isOpened() == False):
    print("Error opening video stream or file")

  frame_width  = int(cap.get(3))
  frame_height = int(cap.get(4))
  nbFrames     = int(cap.get(7))
  inputFps     = int(cap.get(5))

  colorModifTab = [{"red": random.randrange(255), "green": random.randrange(255), "blue": random.randrange(255)} for i in range(1, hyperparameters["nbAnimalsPerWell"])]
  colorModifTab.insert(0, {"red": 0, "green": 0, "blue": 0})
  infoFrame = calculateInfoFrame(superStruct, hyperparameters, nbFrames, colorModifTab)
            
  # Going through the video and printing stuff on it.
  outputFps = inputFps
  if hyperparameters["outputValidationVideoFps"] > 0:
    outputFps = int(hyperparameters["outputValidationVideoFps"])
  
  out = cv2.VideoWriter(outputName,cv2.VideoWriter_fourcc('M','J','P','G'), outputFps, (int(frame_width * hyperparameters["reduceImageResolutionPercentage"]), int(frame_height * hyperparameters["reduceImageResolutionPercentage"])))
  
  if int(hyperparameters["onlyDoTheTrackingForThisNumberOfFrames"]) != 0:
    lastFrame = min(lastFrame, firstFrame + int(hyperparameters["onlyDoTheTrackingForThisNumberOfFrames"]))
  
  cap.set(1, firstFrame)
  
  for l in range(firstFrame, lastFrame):
  
    if l < nbFrames:
    
      if (hyperparameters["freqAlgoPosFollow"] != 0) and (l % hyperparameters["freqAlgoPosFollow"] == 0):
        print("Validation video creation: frame:", l)
    
      ret, frame = cap.read()
      if not ret:
        continue

      frame = processFrame(frame, hyperparameters, infoFrame[l], colorModifTab)
      out.write(frame)
  
  out.release()
  
  if (hyperparameters["freqAlgoPosFollow"] != 0):
    print("Validation video created")
  if hyperparameters["popUpAlgoFollow"]:
    import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow

    popUpAlgoFollow.prepend("Create Validation Video")

  return infoFrame


def createSmallValidationVideosForFlagged(resultFolderPath, offset):
  videoName = os.path.basename(resultFolderPath)
  expectedResultsFile = os.path.join(resultFolderPath, 'results_%s.txt' % videoName)
  if os.path.exists(expectedResultsFile):
    resultsFile = expectedResultsFile
  else:
    if not os.path.exists(resultFolderPath):
      raise ValueError('path %s does not exist' % resultFolderPath)
    resultsFile = next((f for f in os.listdir(resultFolderPath) if os.path.isfile(os.path.join(resultFolderPath, f)) and f.startswith('results_')), None)
    if resultsFile is None:
      raise ValueError('folder %s does not contain the results file' % resultFolderPath)
    resultsFile = os.path.join(resultFolderPath, resultsFile)
  with open(resultsFile) as f:
    supstruct = json.load(f)
  wellPositions = [(max(well["topLeftX"], 0), max(well["topLeftY"], 0), well["lengthX"], well["lengthY"]) for well in supstruct["wellPositions"]]
  flaggedBouts = [(wellIdx, animalIdx, boutIdx, bout["BoutStart"] - offset, bout["BoutEnd"] + offset) for wellIdx, well in enumerate(supstruct["wellPoissMouv"]) for animalIdx, animal in enumerate(well) for boutIdx, bout in enumerate(animal) if bout.get("flag", False)]

  expectedVideoFile = os.path.join(resultFolderPath, '%s.avi' % videoName)
  if os.path.exists(expectedVideoFile):
    videoPath = expectedVideoFile
  else:
    videoFile = next((f for f in os.listdir(resultFolderPath) if os.path.isfile(os.path.join(resultFolderPath, f)) and f.endswith('.avi')), None)
    if videoFile is None:
      raise ValueError('folder %s does not contain the validation video' % resultFolderPath)
    videoPath = os.path.join(resultFolderPath, videoFile)
  cap = zzVideoReading.VideoCapture(videoPath)
  if not cap.isOpened():
    raise ValueError("could not open video file %s" % videoPath)

  outputDirectory = os.path.join(resultFolderPath, 'flaggedBouts')
  if not os.path.exists(outputDirectory):
    os.mkdir(outputDirectory)
  for subvideoIdx, (wellIdx, animalIdx, boutIdx, firstFrame, lastFrame) in enumerate(flaggedBouts):
    frameIdx = max(0, firstFrame)
    x, y, width, height = wellPositions[wellIdx]
    writer = cv2.VideoWriter(os.path.join(outputDirectory, '%s_well%d_animal%d_bout%d.avi' % (videoName, wellIdx, animalIdx, boutIdx)), cv2.VideoWriter_fourcc('M','J','P','G'), 10, (width, height))
    cap.set(1, frameIdx)
    while cap.isOpened() and frameIdx < lastFrame:
      frameIdx += 1
      ret, frame = cap.read()
      if not ret:
        continue
      writer.write(frame[y:y+height,x:x+width])
    writer.release()
  cap.release()
  print('No flagged bouts found in the results.' if not flaggedBouts else 'Subvideos created in %s' % outputDirectory)



import math

import cv2
import numpy as np
from numpy import linspace
from scipy.interpolate import interp1d
from scipy.interpolate import splprep, splev
from scipy.optimize import curve_fit

from ._tailTrackingBase import TailTrackingBase


class HeadEmbeddedTailTrackingMixin(TailTrackingBase):
  @staticmethod
  def __curvatureToXYPositions(successiveAngles, firstAngle, firstX, firstY, distance):
    l    = len(successiveAngles) + 1
    xPos = np.zeros(l)
    yPos = np.zeros(l)
    xPos[0] = firstX
    yPos[0] = firstY
    currentAngle = firstAngle
    for i in range(0, len(successiveAngles)):
      currentAngle = currentAngle + successiveAngles[i]
      xPos[i + 1]  = xPos[i] + distance * math.cos(currentAngle)
      yPos[i + 1]  = yPos[i] + distance * math.sin(currentAngle)
    return [xPos, yPos]

  @staticmethod
  def __sin3Combination(x, a1, a2, a3, b1, b2, b3, c1, c2, c3):
    return a1 * np.sin(a2 * (x - a3)) + b1 * np.sin(b2 * (x - b3)) + c1 * np.sin(c2 * (x - c3))

  @staticmethod
  def __sin2Combination(x, a1, a2, a3, b1, b2, b3):
    return a1 * np.sin(a2 * (x - a3)) + b1 * np.sin(b2 * (x - b3))

  def __smoothBasedOnCurvature(self, points, polynomialDegree):
    from zebrazoom.code.extractParameters import calculateTailAngle

    tailX = points[0]
    tailY = points[1]

    l = len(tailX)
    curvature = np.zeros(l-2)
    distanceC = np.zeros(l-2)
    av = 0
    firstAngle = self._calculateAngle(np.array([tailX[0], tailY[0]]), np.array([tailX[1], tailY[1]]))
    for ii in range(1, l-1):
      angleBef = self._calculateAngle(np.array([tailX[ii-1], tailY[ii-1]]), np.array([tailX[ii],   tailY[ii]]))
      angleAft = self._calculateAngle(np.array([tailX[ii],   tailY[ii]]),   np.array([tailX[ii+1], tailY[ii+1]]))
      curvature[ii-1] = calculateTailAngle(angleBef, angleAft)
      distanceC[ii-1] = math.sqrt((tailX[ii-1] - tailX[ii])**2 + (tailY[ii-1] - tailY[ii])**2)

    if False:
      x = np.linspace(0, len(curvature)-1, len(curvature))
      print("x:", x)
      print("curvature:", curvature)
      popt, pcov = curve_fit(self.__sin2Combination, x, curvature, maxfev=2000)
      curvaturePoly = self.__sin2Combination(np.linspace(0, len(curvature)-1, len(curvature)+1), *popt)
    else:
      x = np.linspace(0,len(curvature)-1,len(curvature))
      # curvature[0] = 0
      poly = np.polyfit(x, curvature, deg=polynomialDegree)
      curvaturePoly = np.polyval(poly, np.linspace(0, len(curvature)-1, len(curvature)+1))
      # curvaturePoly[0] = 0

    [xPosT, yPosT] = self.__curvatureToXYPositions(-curvaturePoly, firstAngle, tailX[0], tailY[0], np.mean(distanceC))

    if False:
      import matplotlib.pyplot as plt

      plt.plot(curvature)
      plt.plot(curvaturePoly)
      plt.show()

      ax = plt.gca()
      ax.invert_yaxis()
      plt.plot(tailX, tailY)
      plt.plot(xPosT, yPosT)
      plt.show()

    return [xPosT, yPosT]

  @staticmethod
  def __smoothTail(points, nbTailPoints, smoothingFactor):
    points3 = [np.array(points[0]), np.array(points[1])]

    y = points[0]
    x = linspace(0, 1, len(y))

    if len(x) > 3:
      interpolated_points = {}

      try:
        if smoothingFactor != -1:
          tck, u = splprep(points3, s=smoothingFactor)
        else:
          tck, u = splprep(points3)
        u2 = [i/(nbTailPoints-1) for i in range(0, nbTailPoints)]
        interpolated_points = splev(u2, tck)
        newX = interpolated_points[0]
        newY = interpolated_points[1]
      except:
        newX = points[0]
        newY = points[1]

    else:
      newX = points[0]
      newY = points[1]

    return [newX, newY]

  @staticmethod
  def __interpolateTail(points, nbTailPoints):

    y = points[0]
    x = linspace(0, 1, len(y))

    if len(x) > 3:

      points2 = []
      for i in range(0, len(points[0])-1):
        if not((points[0][i] == points[0][i+1]) and (points[1][i] == points[1][i+1])):
          points2.append([points[0][i], points[1][i]])

      i = len(points[0]) - 1
      if not((points[0][i-1] == points[0][i]) and (points[1][i-1] == points[1][i])):
        points2.append([points[0][i], points[1][i]])

      points = np.array(points2).T

      # Define some points:
      points = np.array([points[0], points[1]]).T  # a (nbre_points x nbre_dim) array

      # Linear length along the line:
      distance = np.cumsum( np.sqrt(np.sum( np.diff(points, axis=0)**2, axis=1 )) )
      distance = np.insert(distance, 0, 0)/distance[-1]

      # Interpolation for different methods:
      interpolation_method = 'quadratic'    # 'slinear', 'quadratic', 'cubic'
      alpha = np.linspace(0, 1, nbTailPoints)

      interpolated_points = {}

      interpolator =  interp1d(distance, points, kind=interpolation_method, axis=0)
      interpolated_points = interpolator(alpha)

      interpolated_points = interpolated_points.T

      newX = interpolated_points[0]
      newY = interpolated_points[1]

    else:

      newX = points[0]
      newY = points[1]

    return [newX, newY]

  def __findNextPoints(self, depth, x, y, frame, points, angle, maxDepth, steps, nbList, initialImage, debug, dontChooseThisPoint = [], maxRadiusForDontChoosePoint = 0):
    lenX = len(frame[0]) - 1
    lenY = len(frame) - 1

    thetaDiffAccept = 1

    if depth < self._hyperparameters["initialTailPortionMaxSegmentDiffAngleCutOffPos"] * maxDepth:
      thetaDiffAccept = self._hyperparameters["initialTailPortionMaxSegmentDiffAngleValue"]

    if depth > 0.85*maxDepth:
      thetaDiffAccept = 0.6

    if self._hyperparameters["headEmbededMaxAngleBetweenSubsequentSegments"]:
      thetaDiffAccept = self._hyperparameters["headEmbededMaxAngleBetweenSubsequentSegments"]

    pixTotMax = 1000000
    maxTheta  = angle

    l = [i*(math.pi/nbList) for i in range(0,2*nbList) if self._distBetweenThetas(i*(math.pi/nbList), angle) < thetaDiffAccept]

    # if debug:
      # print("debug")

    for step in steps:

      if (step < maxDepth - depth) or (step == steps[0]):

        for theta in l:

          xNew = self._assignValueIfBetweenRange(int(x + step * (math.cos(theta))), 0, lenX)
          yNew = self._assignValueIfBetweenRange(int(y + step * (math.sin(theta))), 0, lenY)
          pixTot = frame[yNew][xNew]

          # if debug:
            # print([theta,pixTot])

          # Keeps that theta angle as maximum if appropriate
          if (pixTot < pixTotMax):
            if (len(dontChooseThisPoint) == 0):
              pixTotMax = pixTot
              maxTheta = theta
              xTot = xNew
              yTot = yNew
            else:
              dist = 1000000000000000000
              for num in range(0, len(dontChooseThisPoint[0])):
                dist = min(dist, math.sqrt((xNew - dontChooseThisPoint[0, num])**2 + (yNew - dontChooseThisPoint[1, num])**2))
              if (not(dist <= maxRadiusForDontChoosePoint)):
                pixTotMax = pixTot
                maxTheta = theta
                xTot = xNew
                yTot = yNew

    w = 4
    ym = yTot - w
    yM = yTot + w
    xm = xTot - w
    xM = xTot + w
    if ym < 0:
      ym = 0
    if xm < 0:
      xm = 0
    if yM > len(initialImage):
      yM = len(initialImage)
    if xM > len(initialImage[0]):
      xM = len(initialImage[0])

    pixSur = np.min(frame[ym:yM, xm:xM]) #initialImage[ym:yM, xm:xM])
    # if debug:
      # print("depth:", depth, " ; maxDepth:", maxDepth, " ; pixSur:", pixSur)

    # if depth > 0.95*maxDepth:
      # pixTot = frame[y][x]
      # if (pixTot < pixTotMax):
        # pixTotMax = pixTot
        # maxTheta = theta
        # xTot = x
        # yTot = y
        # depth = maxDepth + 10

    # if debug:
      # print(["max:",maxTheta,pixTotMax])

    # Calculates distance between new and old point
    distSubsquentPoints = math.sqrt((xTot - x)**2 + (yTot - y)**2)

    pixSurMax = self._hyperparameters["headEmbededParamTailDescentPixThreshStop"]
    # pixSurMax = 220 #150 #245 #150
    if depth + distSubsquentPoints < maxDepth and ((pixSur < pixSurMax) or (depth < self._hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth)):
      points = self._appendPoint(xTot, yTot, points)
    else:
      vectX = xTot - x
      vectY = yTot - y
      xTot  = int(x + (maxDepth / (depth + distSubsquentPoints)) * vectX)
      yTot  = int(y + (maxDepth / (depth + distSubsquentPoints)) * vectY)
      points = self._appendPoint(xTot, yTot, points)
    if debug:
      cv2.circle(frame, (xTot, yTot), 3, (255,0,0),   -1)
      self._debugFrame(frame, title='HeadEmbeddedTailTracking')

    newTheta = self._calculateAngle(x,y,xTot,yTot)
    if distSubsquentPoints > 0 and depth + distSubsquentPoints < maxDepth and ((pixSur < pixSurMax) or (depth < self._hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth)):
      (points,nop) = self.__findNextPoints(depth+distSubsquentPoints,xTot,yTot,frame,points,newTheta,maxDepth,steps,nbList,initialImage,debug)

    if depth == 0:
      lenPoints = len(points[0]) - 1
      if points[0, lenPoints-1] == points[0, lenPoints] and points[1, lenPoints-1] == points[1, lenPoints]:
        points = points[:, :len(points[0])-1]

    return (points,newTheta)

  @staticmethod
  def __weirdTrackingPoints(points, headPosition, tailTip):

    newTrackedTipX = points[0, len(points[0]) - 1]
    newTrackedTipY = points[1, len(points[0]) - 1]

    distInitialHeadToNewTrackedTip = math.sqrt((headPosition[0] - newTrackedTipX)**2 + (headPosition[1] - newTrackedTipY)**2)
    distInitialTipToNewTrackTip    = math.sqrt((tailTip[0] - newTrackedTipX)**2 + (tailTip[1] - newTrackedTipY)**2)

    initialTailLength = math.sqrt((headPosition[0] - tailTip[0])**2 + (headPosition[1] - tailTip[1])**2)

    if (initialTailLength * 0.7 < distInitialHeadToNewTrackedTip) and (distInitialHeadToNewTrackedTip <initialTailLength * 1.3 ) and (distInitialTipToNewTrackTip < initialTailLength):
      return False
    else:
      return True

  def __retrackIfWeirdInitialTracking(self, points, headPosition, tailTip, x, y, frame, angle, maxDepth, nbList, initialImage, i):
    steps = self._hyperparameters["step"]

    if self.__weirdTrackingPoints(points, headPosition, tailTip):
      dontTakeThesePoints       = points.copy()
      dontTakeThesePointsAdding = points.copy()
      pointNumTest   = 0
      steps3 = [indStep for indStep in range(steps[0], steps[len(steps)-1])]
      # First Attempt: for each of the previously tracked points, prevent new tracking from choosing that previously tracked point + steps change
      while (self.__weirdTrackingPoints(points, headPosition, tailTip)) and (pointNumTest < len(dontTakeThesePoints[0])):
        points = np.zeros((2, 0))
        (points, lastFirstTheta2) = self._findNextPoints(0,x,y,frame,points,angle,maxDepth,steps3,nbList,initialImage,self._hyperparameters["debugHeadEmbededFindNextPoints"], np.transpose(np.array([dontTakeThesePoints[:,pointNumTest]])))
        dontTakeThesePointsAdding = np.concatenate((dontTakeThesePointsAdding, points), axis=1)
        dontTakeThesePointsAdding = np.unique(dontTakeThesePointsAdding, axis=1)
        points = np.insert(points, 0, headPosition, axis=1)
        pointNumTest = pointNumTest + 1

      if (pointNumTest == len(dontTakeThesePoints[0])) and (self.__weirdTrackingPoints(points, headPosition, tailTip)):
        pointNumTest   = 0
        # Second Attempt: for each of the previously tracked points, prevent new tracking from choosing any point in a 2 pixel radius of that previously tracked point + steps change
        while (self.__weirdTrackingPoints(points, headPosition, tailTip)) and (pointNumTest < len(dontTakeThesePoints[0])):
          points = np.zeros((2, 0))
          (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,angle,maxDepth,steps3,nbList,initialImage,self._hyperparameters["debugHeadEmbededFindNextPoints"], np.transpose(np.array([dontTakeThesePoints[:,pointNumTest]])), 2)
          dontTakeThesePointsAdding = np.concatenate((dontTakeThesePointsAdding, points), axis=1)
          dontTakeThesePointsAdding = np.unique(dontTakeThesePointsAdding, axis=1)
          points = np.insert(points, 0, headPosition, axis=1)
          pointNumTest = pointNumTest + 1

      if (pointNumTest == len(dontTakeThesePoints[0])) and (self.__weirdTrackingPoints(points, headPosition, tailTip)):
        points = np.zeros((2, 0))
        # Third attempt: prevents new tracking from choosing any of the initially tracked points as well as all the points tracked in the second and third attempt + step change
        (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,angle,maxDepth,steps3,nbList,initialImage,self._hyperparameters["debugHeadEmbededFindNextPoints"], dontTakeThesePointsAdding, 0)
        points = np.insert(points, 0, headPosition, axis=1)

      if (pointNumTest == len(dontTakeThesePoints[0])) and (self.__weirdTrackingPoints(points, headPosition, tailTip)):
        points = np.zeros((2, 0))
        # Third attempt: prevents new tracking from choosing any of the initially tracked points as well as all the points tracked in the second and third attempt + extended step change
        steps2 = [indStep for indStep in range(max(0, steps[0]-1), steps[len(steps)-1]+4)]
        (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,angle,maxDepth,steps2,nbList,initialImage,self._hyperparameters["debugHeadEmbededFindNextPoints"], dontTakeThesePointsAdding, 0)
        points = np.insert(points, 0, headPosition, axis=1)

      if (pointNumTest == len(dontTakeThesePoints[0])) and (self.__weirdTrackingPoints(points, headPosition, tailTip)):
        print("PROBLEM for frame", i, "despite applying correction procedure")
      else:
        print("Ok! Problem solved for frame", i)

    return points

  def _headEmbededTailTracking(self, headPosition, i, frame, maxDepth, tailTip):
    steps   = self._hyperparameters["step"]
    nbList  = 10 if self._hyperparameters["nbList"] == -1 else self._hyperparameters["nbList"]

    x = headPosition[0]
    y = headPosition[1]

    initialImage = frame.copy()

    gaussian_blur = self._hyperparameters["headEmbededParamGaussianBlur"]

    frame = cv2.GaussianBlur(frame, (gaussian_blur, gaussian_blur), 0)

    # angle = self._hyperparameters["headEmbededParamInitialAngle"]
    angle = self._calculateAngle(x, y, tailTip[0], tailTip[1])

    points = np.zeros((2, 0))

    (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,angle,maxDepth,steps,nbList,initialImage,self._hyperparameters["debugHeadEmbededFindNextPoints"])
    points = np.insert(points, 0, headPosition, axis=1)

    # Anomalie detection here
    headEmbededRetrackIfWeirdInitialTracking = self._hyperparameters["headEmbededRetrackIfWeirdInitialTracking"]
    if headEmbededRetrackIfWeirdInitialTracking:
      points = self.__retrackIfWeirdInitialTracking(points, headPosition, tailTip, x, y, frame, angle, maxDepth, nbList, initialImage, i)

    if len(points[0]) > 3:
      if self._hyperparameters["smoothTailHeadEmbeded"]:
        for smoothTailIteration in range(0, self._hyperparameters["smoothTailHeadEmbededNbOfIterations"]):
          points = self.__smoothTail(points, self._nbTailPoints, self._hyperparameters["smoothTailHeadEmbeded"])
      else:
        if not(self._hyperparameters["adjustHeadEmbededTracking"]):
          points   = self._nbTailPoints(points, self._nbTailPoints)
        else:
          nDist    = len(points[0]) - 1
          stepDist = 10 / nDist
          tab      = [int((i/9)*nDist) for i in range(0,10)]
          points   = [[points[0][i] for i in tab], [points[1][i] for i in tab]]


    if False:
      polynomialDegree = 5
      points = self.__smoothBasedOnCurvature(points, polynomialDegree)


    output = np.zeros((1, len(points[0]), 2))

    for idx, x in enumerate(points[0]):
      output[0][idx][0] = x
      output[0][idx][1] = points[1][idx]

    return output

  def _headEmbededTailTrackFindMaxDepth(self, frame):
    if True:
      return math.sqrt((self._headPositionFirstFrame[0] - self._tailTipFirstFrame[0])**2 + (self._headPositionFirstFrame[1] - self._tailTipFirstFrame[1])**2)

    headEmbededParamTailDescentPixThreshStopInit = self._hyperparameters["headEmbededParamTailDescentPixThreshStop"]
    self._hyperparameters["headEmbededParamTailDescentPixThreshStop"] = 256

    x = self._headPositionFirstFrame[0]
    y = self._headPositionFirstFrame[1]

    steps   = self._hyperparameters["step"]
    nbList  = 10

    initialImage = frame.copy()

    gaussian_blur = self._hyperparameters["headEmbededParamGaussianBlur"]

    frame = cv2.GaussianBlur(frame, (gaussian_blur, gaussian_blur), 0)

    angle = self._calculateAngle(x, y, self._tailTipFirstFrame[0], self._tailTipFirstFrame[1])

    points = np.zeros((2, 0))

    (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,angle,self._hyperparameters["headEmbededTailTrackFindMaxDepthInitialMaxDepth"],steps,nbList,initialImage, self._hyperparameters["debugHeadEmbededFindNextPoints"])

    distToTip        = np.full((200),10000)
    distToBase       = np.full((200),10000)
    curTailLengthTab = np.full((200),10000)
    curTailLength  = 0

    k = 0
    distToTip[k]        = abs(math.sqrt((points[0,k]-self._tailTipFirstFrame[0])**2 + (points[1,k]-self._tailTipFirstFrame[1])**2))
    distToBase[k]       = abs(math.sqrt((points[0,k] - x)**2 + (points[1,k] - y)**2))
    curTailLength       = abs(math.sqrt((points[0,k] - x)**2 + (points[1,k] - y)**2))
    curTailLengthTab[k] = curTailLength

    k = 1
    distFromHeadToTip = abs(math.sqrt((x-self._tailTipFirstFrame[0])**2 + (y-self._tailTipFirstFrame[1])**2))
    while (curTailLength < 1.5*distFromHeadToTip) and (k < len(points[0])-1):
      curTailLength = curTailLength + abs(math.sqrt((points[0,k]-points[0,k-1])**2 + (points[1,k]-points[1,k-1])**2))
      distToTip[k]  = abs(math.sqrt((points[0,k]-self._tailTipFirstFrame[0])**2 + (points[1,k]-self._tailTipFirstFrame[1])**2))
      distToBase[k] = abs(math.sqrt((points[0,k] - x)**2 + (points[1,k] - y)**2))
      curTailLengthTab[k] = curTailLength
      k = k + 1

    minDistToTip    = 1000000
    indMinDistToTip = 0
    for idx, dist in enumerate(distToTip):
      if dist < minDistToTip:
        minDistToTip = dist
        indMinDistToTip = idx

    self._hyperparameters["headEmbededParamTailDescentPixThreshStop"] = headEmbededParamTailDescentPixThreshStopInit

    pathFactor = curTailLengthTab[indMinDistToTip] / distToBase[indMinDistToTip]

    return (math.sqrt((x - self._tailTipFirstFrame[0])**2 + (y - self._tailTipFirstFrame[1])**2)* pathFactor)

  def _adjustHeadEmbededHyperparameters(self, frame):
    dist = math.sqrt((self._headPositionFirstFrame[0] - self._tailTipFirstFrame[0])**2 + (self._headPositionFirstFrame[1] - self._tailTipFirstFrame[1])**2)
    factor = dist / 220

    if self._hyperparameters["headEmbededAutoSet_BackgroundExtractionOption"]:
      self._hyperparameters["headEmbededRemoveBack"] = 1
      self._hyperparameters["minPixelDiffForBackExtract"] = self._hyperparameters["headEmbededAutoSet_BackgroundExtractionOption"] # 7
      # if self._hyperparameters["invertBlackWhiteOnImages"] == 0:
      self._hyperparameters["extractBackWhiteBackground"] = 0
      # else:
        # self._hyperparameters["extractBackWhiteBackground"] = 1

    if self._hyperparameters["headEmbededAutoSet_ExtendedDescentSearchOption"]:
      initialStepsTab = [10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50]
    else:
      initialStepsTab = [10, 13, 15]
    self._hyperparameters["step"] = [int(val*factor) for val in initialStepsTab]
    self._hyperparameters["step"] = list(dict.fromkeys(self._hyperparameters["step"])) # removes duplicate values
    if len(self._hyperparameters["step"]) == 0:
      self._hyperparameters["step"] = [2, 4]

    if self._hyperparameters["overwriteFirstStepValue"] or self._hyperparameters["overwriteLastStepValue"]:
      self._hyperparameters["overwriteFirstStepValue"] = int(self._hyperparameters["overwriteFirstStepValue"])
      self._hyperparameters["overwriteLastStepValue"]  = int(self._hyperparameters["overwriteLastStepValue"])
      if self._hyperparameters["overwriteLastStepValue"] <= self._hyperparameters["overwriteFirstStepValue"]:
        self._hyperparameters["overwriteLastStepValue"] = self._hyperparameters["overwriteFirstStepValue"] + 1
      self._hyperparameters["overwriteNbOfStepValues"] = self._hyperparameters["overwriteLastStepValue"] - self._hyperparameters["overwriteFirstStepValue"] + 1
      step = [self._hyperparameters["overwriteFirstStepValue"] + stepVal for stepVal in range(0, self._hyperparameters["overwriteNbOfStepValues"])]
      self._hyperparameters["step"] = step

    self._hyperparameters["headEmbededParamGaussianBlur"] = int(13 * factor)

    if self._hyperparameters["overwriteHeadEmbededParamGaussianBlur"]:
      self._hyperparameters["headEmbededParamGaussianBlur"] = int(self._hyperparameters["overwriteHeadEmbededParamGaussianBlur"])

    if self._hyperparameters["headEmbededParamGaussianBlur"] % 2 == 0:
      self._hyperparameters["headEmbededParamGaussianBlur"] = self._hyperparameters["headEmbededParamGaussianBlur"] + 1

    self._hyperparameters["addBlackCircleOfHalfDiamOnHeadForBoutDetect"] = int(70 * factor)

    self._hyperparameters["headEmbededTailTrackFindMaxDepthInitialMaxDepth"] = int(300*factor)

    self._hyperparameters["trackingPointSizeDisplay"] = int(5*factor) if int(5*factor) > 0 else 1 # CAN DO BETTER HERE !!!

    tailTipSurroundingDiameter = int(15 * factor)
    x, y = self._tailTipFirstFrame
    videoWidth  = self._hyperparameters["videoWidth"]
    videoHeight = self._hyperparameters["videoHeight"]
    if y-tailTipSurroundingDiameter < 0:
      ymin = 0
    else:
      ymin = y-tailTipSurroundingDiameter
    if y+tailTipSurroundingDiameter > videoHeight:
      ymax = videoHeight - 1
    else:
      ymax = y+tailTipSurroundingDiameter
    if x-tailTipSurroundingDiameter < 0:
      xmin = 0
    else:
      xmin = x-tailTipSurroundingDiameter
    if x+tailTipSurroundingDiameter > videoWidth:
      xmax = videoWidth - 1
    else:
      xmax = x+tailTipSurroundingDiameter
    self._hyperparameters["headEmbededParamTailDescentPixThreshStop"] = int(np.mean(frame[int(ymin):int(ymax), int(xmin):int(xmax)]))

    if self._hyperparameters["headEmbededParamTailDescentPixThreshStopOverwrite"] != -1:
      self._hyperparameters["headEmbededParamTailDescentPixThreshStop"] = self._hyperparameters["headEmbededParamTailDescentPixThreshStopOverwrite"]

    if False:
      print("headEmbededParamTailDescentPixThreshStop:", self._hyperparameters["headEmbededParamTailDescentPixThreshStop"])
      print("dist:", dist)
      print("step:", self._hyperparameters["step"])
      print("headEmbededParamGaussianBlur:", self._hyperparameters["headEmbededParamGaussianBlur"])
      print("trackingPointSizeDisplay:", self._hyperparameters["trackingPointSizeDisplay"])



import contextlib
import copy
import itertools
import json
import os

import pandas as pd

from PyQt5.QtCore import pyqtSignal, Qt, QAbstractTableModel, QDir, QEvent, QItemSelection, QItemSelectionModel, QModelIndex, QObject, QPoint, QPointF, QRect, QRectF, QSize, QSizeF, QSortFilterProxyModel, QUrl
from PyQt5.QtGui import QColor, QDesktopServices, QFont, QPainter, QPixmap, QPolygon, QPolygonF, QTransform
from PyQt5.QtWidgets import QAbstractItemView, QApplication, QFileDialog, QFileSystemModel, QFrame, QGraphicsPixmapItem, QGraphicsScene, QGraphicsView, QHBoxLayout, QHeaderView, QLabel, QListView, QMessageBox, QPushButton, QScrollArea, QSlider, QSpacerItem, QStyleOptionSlider, QTableView, QTextEdit, QTreeView, QToolTip, QVBoxLayout, QWidget

import zebrazoom.code.paths as paths
import zebrazoom.code.util as util
from zebrazoom.code.GUI.GUI_InitialClasses import getVideosFromResultsGroups
from zebrazoom.code.GUI.readValidationVideo import getFramesCallback


class RolloverAnalysis(QWidget):
  def __init__(self, controller):
    super().__init__(controller.window)
    self.controller = controller
    self.preferredSize = (900, 450)

    layout = QVBoxLayout()
    layout.addWidget(util.apply_style(QLabel("Zebrafish rollover analysis", self), font=controller.title_font), alignment=Qt.AlignmentFlag.AlignCenter)

    classificationBtn = util.apply_style(QPushButton("Manually classify video frames into rollover vs no-rollover", self), background_color=util.LIGHT_YELLOW)
    classificationBtn.clicked.connect(lambda: controller.showManualRolloverClassification())
    layout.addWidget(classificationBtn, alignment=Qt.AlignmentFlag.AlignCenter)

    openConfigurationsFolderBtn = util.apply_style(QPushButton("Open rollover detection configuration file folder"), background_color=util.LIGHT_YELLOW)
    openConfigurationsFolderBtn.clicked.connect(lambda: QDesktopServices.openUrl(QUrl.fromLocalFile(paths.getRolloverDetectionConfigurationFolder())))
    layout.addWidget(openConfigurationsFolderBtn, alignment=Qt.AlignmentFlag.AlignCenter)

    oneVideoDetectionBtn = util.apply_style(QPushButton("Launch rollover detection on one video", self), background_color=util.LIGHT_YELLOW)
    oneVideoDetectionBtn.clicked.connect(_ResultsSelectionPage.runSingle)
    layout.addWidget(oneVideoDetectionBtn, alignment=Qt.AlignmentFlag.AlignCenter)

    multipleVideosDetectionBtn = util.apply_style(QPushButton("Launch rollover detection on multiple videos", self), background_color=util.LIGHT_YELLOW)
    multipleVideosDetectionBtn.clicked.connect(_showTemporaryPage)
    layout.addWidget(multipleVideosDetectionBtn, alignment=Qt.AlignmentFlag.AlignCenter)

    comparePopulationsBtn = util.apply_style(QPushButton("Compare zebrafish populations based on rollover detection", self), background_color=util.LIGHT_YELLOW)
    comparePopulationsBtn.clicked.connect(lambda: controller.show_frame("CreateExperimentOrganizationExcel"))
    layout.addWidget(comparePopulationsBtn, alignment=Qt.AlignmentFlag.AlignCenter)

    documentationBtn = util.apply_style(QPushButton("Read rollover detection documentation", self), background_color=util.LIGHT_GREEN)
    documentationBtn.clicked.connect(lambda: QDesktopServices.openUrl(QUrl("https://zebrazoom.org/documentation/docs/behaviorAnalysis/zebrafishRolloverDetection")))
    layout.addWidget(documentationBtn, alignment=Qt.AlignmentFlag.AlignCenter)

    start_page_btn = QPushButton("Go to the start page", self)
    start_page_btn.clicked.connect(lambda: controller.show_frame("StartPage"))
    layout.addWidget(start_page_btn, alignment=Qt.AlignmentFlag.AlignCenter)

    self.setLayout(layout)


class _ConfigurationsModel(QFileSystemModel):
  def __init__(self):
    super().__init__()
    self.setReadOnly(False)
    self.setNameFilterDisables(False)
    self.setNameFilters(("*.csv",))

  def data(self, index, role=Qt.ItemDataRole.DisplayRole):
    data = super().data(index, role=role)
    if role == Qt.ItemDataRole.EditRole:
      return os.path.splitext(data)[0]
    return data

  def setData(self, index, value, role=None):
    extension = os.path.splitext(self.data(index))[1]
    return super().setData(index, "%s%s" % (value, extension), role=role)


class _ConfigurationsTreeView(QTreeView):
  def __init__(self):
    super().__init__()
    self.sizeHint = lambda: QSize(150, 1)
    self.header().setSectionResizeMode(QHeaderView.ResizeMode.Fixed)
    self.setRootIsDecorated(False)

  def resizeEvent(self, evt):
    super().resizeEvent(evt)
    self.setColumnWidth(0, evt.size().width())

  def flags(self, index):
    return super().flags(index) | Qt.ItemFlag.ItemIsEditable


class _ResultsModel(QAbstractTableModel):
  _COLUMN_TITLES = ["Video", "Config"]
  _DEFAULT_ZZOUTPUT = paths.getDefaultZZoutputFolder()

  def __init__(self, filename):
    super().__init__()
    data = pd.read_csv(filename).fillna('')
    self._videos = data['Video'].tolist()
    self._configs = data['Config'].tolist()

  def rowCount(self, parent=None):
    return len(self._videos)

  def columnCount(self, parent=None):
    return len(self._COLUMN_TITLES)

  def data(self, index, role=Qt.ItemDataRole.DisplayRole):
    if index.isValid() and role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole, Qt.ItemDataRole.ToolTipRole):
      return self._videos[index.row()] if index.column() == 0 else self._configs[index.row()]
    return None

  def headerData(self, col, orientation, role):
    if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
      return self._COLUMN_TITLES[col]
    return None

  def addVideos(self, videos):
    if videos is None:
      return
    videos = [video for video in videos if video not in self._videos]
    self.beginInsertRows(QModelIndex(), self.rowCount(), self.rowCount() + len(videos) - 1)
    self._videos.extend(videos)
    self._configs.extend([''] * len(videos))
    self.endInsertRows()

  def setConfigs(self, idxs, config):
    if config is None:
      return
    self.beginResetModel()
    for idx in idxs:
      self._configs[idx] = config
    self.endResetModel()

  def removeSelectedRows(self, idxs):
    self.beginResetModel()
    for idx in reversed(sorted(idxs)):
      del self._videos[idx]
      del self._configs[idx]
    self.endResetModel()

  def getData(self):
    return self._videos, self._configs

  def hasUnsavedChanges(self, filename):
    if not os.path.exists(filename):  # file was deleted
      return False
    fileData = pd.read_csv(filename).fillna('')
    return not (fileData['Video'].tolist() == self._videos and fileData['Config'].tolist() == self._configs)

  def saveFile(self, filename):
    pd.DataFrame(columns=_ResultsModel._COLUMN_TITLES, data=zip(self._videos, self._configs)).to_csv(filename, index=False)


class _ConfigurationsSelectionModel(QItemSelectionModel):
  def __init__(self, window, table, *args):
    super().__init__(*args)
    self._window = window
    self._table = table
    self._blockSelection = False

  def setCurrentIndex(self, index, command):
    if index != self.currentIndex() and self._table.model() is not None and self._table.model().hasUnsavedChanges(self.model().filePath(self.currentIndex())) and \
        QMessageBox.question(self._window, "Unsaved changes", "Are you sure you want to proceed? Unsaved changes will be lost.",
                             defaultButton=QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
      self._blockSelection = True
      return None
    return super().setCurrentIndex(index, command)

  def select(self, index, command):
    if self._blockSelection:
      self._blockSelection = False
      return None
    return super().select(index, command)


class _ResultsSelectionPage(QWidget):
  def __init__(self):
    super().__init__()

    app = QApplication.instance()

    layout = QVBoxLayout()
    layout.addWidget(util.apply_style(QLabel("Select results and corresponding rollover config files"), font=app.title_font), alignment=Qt.AlignmentFlag.AlignCenter)

    folderPath = os.path.join(paths.getRootDataFolder(), 'rolloverConfigurations')
    if not os.path.exists(folderPath):
      os.makedirs(folderPath)
    model = _ConfigurationsModel()
    model.setRootPath(folderPath)
    self._tree = tree = _ConfigurationsTreeView()
    tree.setModel(model)
    for idx in range(1, model.columnCount()):
      tree.hideColumn(idx)
    tree.setRootIndex(model.index(model.rootPath()))
    self._table = QTableView()
    selectionModel = _ConfigurationsSelectionModel(app.window, self._table, model)
    tree.setSelectionModel(selectionModel)
    selectionModel.currentRowChanged.connect(lambda current, previous: current.row() == -1 or self._fileSelected(model.filePath(current)))

    treeLayout = QVBoxLayout()
    self._newConfigurationBtn = QPushButton("New configuration")
    self._newConfigurationBtn.clicked.connect(self._newConfiguration)
    treeLayout.addWidget(self._newConfigurationBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    treeLayout.addWidget(tree, stretch=1)
    treeWidget = QWidget()
    treeLayout.setContentsMargins(0, 0, 0, 0)
    treeWidget.setLayout(treeLayout)
    horizontalSplitter = util.CollapsibleSplitter()
    horizontalSplitter.addWidget(treeWidget)

    tableLayout = QVBoxLayout()
    tableButtonsLayout = QHBoxLayout()
    self._addMultipleFoldersBtn = QPushButton("Add results folders")
    self._addMultipleFoldersBtn.clicked.connect(self._addMultipleFolders)
    tableButtonsLayout.addWidget(self._addMultipleFoldersBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    self._addResultsGroupsBtn = QPushButton("Add results groups")
    self._addResultsGroupsBtn.clicked.connect(self._addResultsGroups)
    tableButtonsLayout.addWidget(self._addResultsGroupsBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    chooseConfigsBtn = QPushButton("Choose config for selected results")
    chooseConfigsBtn.clicked.connect(lambda: self._table.model().setConfigs(sorted(set(map(lambda idx: idx.row(), self._table.selectionModel().selectedIndexes()))),
                                                                            QFileDialog.getOpenFileName(app.window, 'Select config file', paths.getRolloverDetectionConfigurationFolder(), "JSON (*.json)")[0]))
    tableButtonsLayout.addWidget(chooseConfigsBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    removeVideosBtn = QPushButton("Remove selected results")
    removeVideosBtn.clicked.connect(lambda: self._table.model().removeSelectedRows(sorted(set(map(lambda idx: idx.row(), self._table.selectionModel().selectedIndexes())))))
    tableButtonsLayout.addWidget(removeVideosBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    saveChangesBtn = QPushButton("Save changes")
    saveChangesBtn.clicked.connect(lambda: self._table.model().saveFile(self._tree.model().filePath(selectionModel.currentIndex())) or QMessageBox.information(app.window, "Configuration saved", "Changes made to the configuration were saved."))
    tableButtonsLayout.addWidget(saveChangesBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    deleteConfigurationBtn = QPushButton("Delete configuration")
    deleteConfigurationBtn.clicked.connect(self._removeConfiguration)
    tableButtonsLayout.addWidget(deleteConfigurationBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    openConfigurationsFolderBtn = QPushButton("Open configurations folder")
    openConfigurationsFolderBtn.clicked.connect(lambda: QDesktopServices.openUrl(QUrl.fromLocalFile(folderPath)))
    tableButtonsLayout.addWidget(openConfigurationsFolderBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    self._runTrackingBtn = util.apply_style(QPushButton("Run tracking"), background_color=util.DEFAULT_BUTTON_COLOR)
    self._runTrackingBtn.clicked.connect(self._unsavedChangesWarning(lambda *_: self._runTracking(), forceSave=True))
    tableButtonsLayout.addWidget(self._runTrackingBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    tableButtonsLayout.addStretch()
    tableLayout.addLayout(tableButtonsLayout)
    tableLayout.addWidget(self._table, stretch=1)

    self._mainWidget = QWidget()
    tableLayout.setContentsMargins(0, 0, 0, 0)
    self._mainWidget.setLayout(tableLayout)

    scrollArea = QScrollArea()
    scrollArea.setFrameShape(QFrame.Shape.NoFrame)
    scrollArea.setWidgetResizable(True)
    scrollArea.setWidget(self._mainWidget)
    horizontalSplitter.addWidget(scrollArea)
    horizontalSplitter.setStretchFactor(1, 1)
    horizontalSplitter.setChildrenCollapsible(False)
    layout.addWidget(horizontalSplitter, stretch=1)

    buttonsLayout = QHBoxLayout()
    buttonsLayout.addStretch()
    startPageBtn = QPushButton("Go to the start page")
    startPageBtn.clicked.connect(self._unsavedChangesWarning(lambda *_: app.show_frame("StartPage")))
    buttonsLayout.addWidget(startPageBtn, alignment=Qt.AlignmentFlag.AlignCenter)
    buttonsLayout.addStretch()
    layout.addLayout(buttonsLayout)

    self.setLayout(layout)
    self._mainWidget.hide()

  @staticmethod
  def _findResultsFile(path):
    if not os.path.exists(path):
      return None
    folder = os.path.basename(path)
    reference = os.path.join(path, 'results_' + folder + '.txt')
    if os.path.exists(reference):
      return reference
    resultsFile = next((f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f)) if f.startswith('results_')), None)
    if resultsFile is None:
      return None
    return os.path.join(path, resultsFile)

  def _getMultipleFolders(self):
    app = QApplication.instance()
    dialog = QFileDialog()
    dialog.setWindowTitle('Select one or more results folders (use Ctrl or Shift key to select multiple folders)')
    dialog.setDirectory(app.ZZoutputLocation)
    dialog.setFileMode(QFileDialog.FileMode.Directory)
    dialog.setOption(QFileDialog.Option.DontUseNativeDialog, True)
    dialog.setOption(QFileDialog.Option.ShowDirsOnly, True)
    listView = dialog.findChild(QListView, 'listView')
    if listView:
      listView.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
    treeView = dialog.findChild(QTreeView)
    if treeView:
      treeView.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

    if not dialog.exec():
      return None
    return dialog.selectedFiles()

  @contextlib.contextmanager
  def __selectAddedRows(self):
    app = QApplication.instance()
    model = self._table.model()
    firstNewRow = model.rowCount()
    yield
    if firstNewRow == model.rowCount():
      return  # no videos added
    addedRows = QItemSelection()
    addedRows.select(model.index(firstNewRow, 0), model.index(model.rowCount() - 1, 0))
    self._table.selectionModel().setCurrentIndex(model.index(firstNewRow, 0), QItemSelectionModel.SelectionFlag.ClearAndSelect | QItemSelectionModel.SelectionFlag.Rows)
    self._table.selectionModel().select(addedRows, QItemSelectionModel.SelectionFlag.ClearAndSelect | QItemSelectionModel.SelectionFlag.Rows)
    self._table.setFocus()

  def __addVideos(self, paths):
    invalidFolders = []
    for path in paths:
      if self._findResultsFile(path) is None:
        invalidFolders.append(path)
        continue
    with self.__selectAddedRows():
      self._table.model().addVideos([os.path.normpath(path) for path in paths if path not in invalidFolders])
    if invalidFolders:
      app = QApplication.instance()
      warning = QMessageBox(app.window)
      warning.setIcon(QMessageBox.Icon.Warning)
      warning.setWindowTitle("Invalid folders selected")
      warning.setText("Some of the selected folders were ignored because they are not valid results folders.")
      warning.setDetailedText("\n".join(invalidFolders))
      warning.exec()

  def _addMultipleFolders(self):
    selectedFolders = self._getMultipleFolders()
    if selectedFolders is None:
      return
    self.__addVideos(selectedFolders)

  def _addResultsGroups(self):
    selectedVideos = getVideosFromResultsGroups(showH5=False)
    if selectedVideos is None:
      return
    self.__addVideos([path for path in selectedVideos if os.path.isdir(path)])  # only legacy results are supported in rollover analysis

  def _unsavedChangesWarning(self, fn, forceSave=False):
    app = QApplication.instance()
    def inner(*args, **kwargs):
      if forceSave:
        text = "Do you want to save the changes and proceed?"
      else:
        text = "Are you sure you want to proceed? Unsaved changes will be lost."
      filename = self._tree.model().filePath(self._tree.selectionModel().currentIndex())
      if self._table.model() is not None and self._table.model().hasUnsavedChanges(filename):
        if QMessageBox.question(app.window, "Unsaved changes", text, defaultButton=QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
          return
        elif forceSave:
          self._table.model().saveFile(filename)
      return fn(*args, **kwargs)
    return inner

  def _newConfiguration(self):
    number = 1
    while os.path.exists(os.path.join(self._tree.model().rootPath(), 'Configuration %d.csv' % number)):
      number += 1
    path = os.path.join(self._tree.model().rootPath(), 'Configuration %d.csv' % number)
    pd.DataFrame(columns=_ResultsModel._COLUMN_TITLES).to_csv(path, index=False)
    index = self._tree.model().index(path)
    self._tree.selectionModel().setCurrentIndex(index, QItemSelectionModel.SelectionFlag.ClearAndSelect)
    self._tree.edit(index)

  def _removeConfiguration(self):
    app = QApplication.instance()
    if QMessageBox.question(app.window, "Delete configuration", "Are you sure you want to delete the configuration? This action removes the file from disk and cannot be undone.",
                            defaultButton=QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
      return
    pathToRemove = self._tree.model().filePath(self._tree.selectionModel().currentIndex())
    self._tree.model().remove(self._tree.selectionModel().currentIndex())
    if pathToRemove == self._tree.model().filePath(self._tree.selectionModel().currentIndex()):  # last valid file removed
      self._mainWidget.hide()

  def _fileSelected(self, filename):
    self._mainWidget.show()
    self._table.setModel(_ResultsModel(filename))
    self._table.horizontalHeader().resizeSections(QHeaderView.ResizeMode.Stretch)

  @staticmethod
  def __run(resultsFolder, config):
    from zzdeeprollover.detectRolloverFrames import detectRolloverFrames
    comparePredictedWithManual = os.path.exists(os.path.join(resultsFolder, 'rolloverManualClassification.json'))
    detectRolloverFrames(os.path.basename(resultsFolder), os.path.dirname(resultsFolder), config['medianRollingMean'], config['resizeCropDimension'], comparePredictedWithManual, 1, config['imagesToClassifyHalfDiameter'], config['modelPath'])

  @util.showInProgressPage('Rollover detection')
  def __runMultiple(self, results, configs):
    app = QApplication.instance()
    for resultsFolder, config in zip(results, configs):
      self.__run(resultsFolder, config)
    QMessageBox.information(app.window, "Rollover detection done", "Rollover detection was completed successfully.")

  @classmethod
  def runSingle(cls):
    app = QApplication.instance()
    resultsFolder = QFileDialog.getExistingDirectory(app.window, 'Select the results folder on which you want to perform rollover detection', app.ZZoutputLocation)
    if not resultsFolder:
      return
    if cls._findResultsFile(resultsFolder) is None:
      QMessageBox.critical(app.window, 'Invalid folder selected', f'Folder {resultsFolder} is not a valid results folder.')
      return

    configPath, _ = QFileDialog.getOpenFileName(app.window, 'Select the configuration file', paths.getRolloverDetectionConfigurationFolder(), 'JSON (*.json)')
    if not configPath:
      return
    with open(configPath) as f:
      config = json.load(f)
    expectedKeys = {'medianRollingMean', 'resizeCropDimension', 'imagesToClassifyHalfDiameter', 'modelPath'}
    if set(config) != expectedKeys:
      QMessageBox.critical(app.window, 'Invalid configuration file selected', f'Configuration file contains invalid keys. Expected keys are {", ".join(expectedKeys)}.')
      return
    if not os.path.isabs(config['modelPath']):
      config['modelPath'] = os.path.join(os.path.dirname(configPath), config['modelPath'])

    util.showInProgressPage('Rollover detection')(cls.__run)(resultsFolder, config)
    app.show_frame('RolloverAnalysis')
    QMessageBox.information(app.window, "Rollover detection done", "Rollover detection was completed successfully.")

  def _runTracking(self):
    app = QApplication.instance()
    results, configs = self._table.model().getData()
    errors = []
    loadedConfigs = []
    for resultsFolder, config in zip(results, configs):
      if not config:
        errors.append('Config is not specified for video %s.' % video)
      elif not os.path.exists(config):
        errors.append('Config %s does not exist.' % config)
      else:
        with open(config) as f:
          loadedConfig = json.load(f)
          loadedConfigs.append(loadedConfig)
        expectedKeys = {'medianRollingMean', 'resizeCropDimension', 'imagesToClassifyHalfDiameter', 'modelPath'}
        if set(loadedConfig) != expectedKeys:
          errors.append(f'Config {config} contains invalid keys. Expected keys are {", ".join(expectedKeys)}.')
        if not os.path.isabs(loadedConfig['modelPath']):
          loadedConfig['modelPath'] = os.path.join(os.path.dirname(config), loadedConfig['modelPath'])
      if not os.path.exists(resultsFolder):
        errors.append(f'Folder {resultsFolder} does not exist.')
      if self._findResultsFile(resultsFolder) is None:
        errors.append(f'Folder {resultsFolder} is not a valid results folder.')
    if errors:
      error = QMessageBox(app.window)
      error.setIcon(QMessageBox.Icon.Critical)
      error.setWindowTitle("Specification contains some errors")
      error.setText("Cannot detect rollovers because the specified results and config combinations contain some errors:")
      error.setDetailedText("\n".join(errors))
      textEdit = error.findChild(QTextEdit)
      textEdit.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)
      layout = error.layout()
      layout.addItem(QSpacerItem(600, 0), layout.rowCount(), 0, 1, layout.columnCount())
      error.exec()
      return
    stackedLayout =  app.window.centralWidget().layout()
    blocked = stackedLayout.blockSignals(True)
    self.__runMultiple(results, loadedConfigs)
    stackedLayout.blockSignals(blocked)


def _showTemporaryPage():
  app = QApplication.instance()
  layout = app.window.centralWidget().layout()
  page = _ResultsSelectionPage()
  layout.addWidget(page)
  layout.setCurrentWidget(page)
  def cleanup():
    layout.removeWidget(page)
    layout.currentChanged.disconnect(cleanup)
  layout.currentChanged.connect(cleanup)


class _TooltipHelper(QObject):
  def eventFilter(self, obj, evt):
    if evt.type() != QEvent.Type.ToolTip:
      return False
    view = obj.parent()
    if view is None:
      return False

    index = view.indexAt(evt.pos())
    if not index.isValid():
      return False
    rect = view.visualRect(index)
    if view.sizeHintForColumn(index.column()) > rect.width():
      QToolTip.showText(evt.globalPos(), view.model().data(index), view, rect)
      return True
    else:
      QToolTip.hideText()
      return True
    return False


class _WellSelectionImage(QGraphicsView):
  wellChanged = pyqtSignal(int)

  def __init__(self):
    super().__init__()
    self._well = None
    self._hoveredWell = None
    self._wellPositions = None
    self.wellShape = None

    self._zoom = 0
    self._scene = QGraphicsScene(self)
    self._pixmap = QGraphicsPixmapItem()
    self._scene.addItem(self._pixmap)
    self.setScene(self._scene)
    self.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
    self.setResizeAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
    self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    self.setFrameShape(QFrame.Shape.NoFrame)
    self._dragging = False

  def fitInView(self):
    rect = QRectF(self._pixmap.pixmap().rect())
    self.setSceneRect(rect)
    unity = self.transform().mapRect(QRectF(0, 0, 1, 1))
    self.scale(1 / unity.width(), 1 / unity.height())
    viewrect = self.viewport().rect()
    scenerect = self.transform().mapRect(rect)
    factor = min(viewrect.width() / scenerect.width(),
                 viewrect.height() / scenerect.height())
    self.scale(factor, factor)
    self._zoom = 0

  def _update(self, scaleFactor):
    if self._zoom > 0:
      self.scale(scaleFactor, scaleFactor)
    elif self._zoom == 0:
      self.fitInView()
    else:
      self._zoom = 0

  def wheelEvent(self, evt):
    if evt.angleDelta().y() > 0:
      self._zoom += 1
      self._update(1.25)
    else:
      self._zoom -= 1
      self._update(0.8)

  def keyPressEvent(self, evt):
    if evt.modifiers() & Qt.KeyboardModifier.ControlModifier:
      if evt.key() == Qt.Key.Key_Plus:
        self._zoom += 1
        self._update(1.25)
        return
      if evt.key() == Qt.Key.Key_Minus:
        self._zoom -= 1
        self._update(0.8)
        return
    super().keyPressEvent(evt)

  def setWellPositions(self, wellPositions):
    self.setMouseTracking(wellPositions is not None)
    self._wellPositions = wellPositions
    self._well = 0
    self.wellChanged.emit(self._well)

  def setOriginalPixmap(self, pixmap, update=True):
    self._pixmap.setPixmap(pixmap)
    if update:
      self._well = 0
      self._zoom = 0
      self.setDragMode(QGraphicsView.DragMode.ScrollHandDrag)
      self.fitInView()

  def mouseMoveEvent(self, evt):
    if evt.buttons() == Qt.MouseButton.LeftButton:
      if not self._dragging:
        self._dragging = True
        QApplication.setOverrideCursor(Qt.CursorShape.ClosedHandCursor)
      super().mouseMoveEvent(evt)
      return

    app = QApplication.instance()
    if not self._wellPositions:
      return
    oldHovered = self._hoveredWell
    if self.wellShape == 'rectangle':
      def test_func(shape, x, y, width, height):
        return QRectF(x, y, width, height).contains(shape)
    else:
      assert self.wellShape == 'circle'
      def test_func(shape, x, y, width, height):
        radius = width / 2
        centerX = x + radius
        centerY = y + radius
        dx = abs(shape.x() - centerX)
        if dx > radius:
          return False
        dy = abs(shape.y() - centerY)
        if dy > radius:
          return False
        if dx + dy <= radius:
          return True
        return dx * dx + dy * dy <= radius * radius
    self._hoveredWell = None
    for idx, positions in enumerate(self._wellPositions):
      point = self.mapToScene(evt.pos())
      if test_func(point, *positions):
        self._hoveredWell = idx
        break
    if self._hoveredWell != oldHovered:
      self.viewport().update()
    super().mouseMoveEvent(evt)

  def mousePressEvent(self, evt):
    if not self._wellPositions or self._hoveredWell is None:
      super().mousePressEvent(evt)
      return
    if self._well != self._hoveredWell:
      self._well = self._hoveredWell
      self.wellChanged.emit(self._well)
      self.viewport().update()
    super().mousePressEvent(evt)

  def mouseReleaseEvent(self, evt):
    if self._dragging:
      self._dragging = False
      QApplication.restoreOverrideCursor()
    super().mouseReleaseEvent(evt)

  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.CrossCursor)
    super().enterEvent(evt)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()
    self._hoveredWell = None
    self.viewport().update()
    super().leaveEvent(evt)

  def paintEvent(self, evt):
    super().paintEvent(evt)
    app = QApplication.instance()
    if not self._wellPositions:
      return
    qp = QPainter(self.viewport())
    factory = qp.drawRect if self.wellShape == 'rectangle' else qp.drawEllipse
    for idx, positions in enumerate(self._wellPositions):
      if idx == self._well:
        qp.setPen(QColor(255, 0, 0))
      elif idx == self._hoveredWell:
        qp.setPen(QColor(0, 255, 0))
      else:
        qp.setPen(QColor(0, 0, 255))
      rect = self.mapFromScene(QRectF(*positions)).boundingRect()
      font = QFont()
      font.setPointSize(16)
      font.setWeight(QFont.Weight.Bold)
      qp.setFont(font)
      if idx != self._well:
        qp.drawText(rect, Qt.AlignmentFlag.AlignCenter, str(idx))
      factory(rect)
    qp.end()

  def getWell(self):
    return self._well if self._wellPositions is not None else 0

  def totalWells(self):
    return len(self._wellPositions) if self._wellPositions is not None else 1


class RolloverSlider(QSlider):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.rolloverRanges = []
    self.inBetweenRanges = []

  def paintEvent(self, event):
    style = self.style()
    opt = QStyleOptionSlider()
    self.initStyleOption(opt)
    available = style.pixelMetric(style.PM_SliderSpaceAvailable, opt, self)
    minimum = self.minimum()
    maximum = self.maximum()
    qp = QPainter(self)
    qp.translate(opt.rect.x(), opt.rect.y())
    for color, frameRange in zip((QColor('red'), QColor('blue')), (self.rolloverRanges, self.inBetweenRanges)):
      for start, end in frameRange:
        start = style.sliderPositionFromValue(minimum, maximum, start, available, opt.upsideDown)
        end = style.sliderPositionFromValue(minimum, maximum, end, available, opt.upsideDown)
        qp.fillRect(start, 0, end - start + 1, event.rect().height(), color)
    super().paintEvent(event)


class ManualRolloverClassification(util.CollapsibleSplitter):
  def __init__(self, controller):
    super().__init__(controller.window)
    self.controller = controller
    self._rolloverClassification = {}

    model = QFileSystemModel()
    model.setFilter(QDir.Filter.NoDotAndDotDot | QDir.Filter.Dirs)
    model.setRootPath(self.controller.ZZoutputLocation)
    model.setReadOnly(True)
    proxyModel = QSortFilterProxyModel()

    def filterModel(row, parent):
      index = model.index(row, 0, parent)
      if os.path.normpath(model.filePath(index)) == os.path.normpath(self.controller.ZZoutputLocation):
        return True
      return bool(self._findResultsFile(model.fileName(index)))
    proxyModel.filterAcceptsRow = filterModel
    proxyModel.setSourceModel(model)
    self._tree = tree = QTreeView()
    tree.viewport().installEventFilter(_TooltipHelper(tree))
    tree.sizeHint = lambda: QSize(150, 1)
    tree.setModel(proxyModel)
    tree.setRootIsDecorated(False)
    tree.header().setSectionResizeMode(QHeaderView.ResizeMode.Fixed)
    for idx in range(1, model.columnCount()):
      tree.hideColumn(idx)
    tree.resizeEvent = lambda evt: tree.setColumnWidth(0, evt.size().width())
    selectionModel = tree.selectionModel()
    selectionModel.currentRowChanged.connect(lambda current, previous: current.row() == -1 or self.setFolder(model.fileName(current)))

    layout = QVBoxLayout()
    self._title_label = util.apply_style(QLabel('', self), font_size='16px')
    layout.addWidget(self._title_label, alignment=Qt.AlignmentFlag.AlignCenter)

    self._frame = frame = _WellSelectionImage()
    frame.wellChanged.connect(self._wellSelected)
    layout.addWidget(frame, stretch=1)
    self._slider = slider = RolloverSlider(Qt.Orientation.Horizontal)
    slider.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
    slider.setPageStep(50)
    self._getFrame = lambda *args: None
    slider.valueChanged.connect(lambda: frame.setOriginalPixmap(QPixmap(util._cvToPixmap(self._getFrame(slider))), update=False))
    layout.addWidget(slider)

    self._messageLabel = QLabel('Video not found for the selected results.')
    layout.addWidget(self._messageLabel, stretch=1)
    self._messageLabel.setVisible(False)

    layout.addWidget(QLabel('<span style="color:red">Rollover ranges are marked with red,</span> <span style="color:blue">in between are marked with blue.</span>'), alignment=Qt.AlignmentFlag.AlignCenter)

    self._currentRange = None
    currentRangeIdx = None
    slider.valueChanged.connect(lambda value: updateRollover(value))
    rolloverLayout = QHBoxLayout()

    def getRange(frame, ranges):
      for idx, (start, end) in enumerate(ranges):
        if start <= frame <= end:
          return (start, end), idx
      return None, None

    def updateRollover(frame):
      nonlocal currentRangeIdx
      if self._currentRange is not None:
        start, end = sorted((self._currentRange[0], frame))
        if any(start <= val <= end for range_ in itertools.chain(slider.rolloverRanges, slider.inBetweenRanges) for val in range_ if range_ is not self._currentRange):
          slider.setValue(self._currentRange[1])
        else:
          self._currentRange[1] = frame
        return
      saveButton.setEnabled(self._originalRolloverClassification != self._rolloverClassification)
      rolloverButton.setEnabled(True)
      inBetweenButton.setEnabled(True)
      rolloverRange, currentRangeIdx = getRange(frame, slider.rolloverRanges)
      if rolloverRange is not None:
        start, end = rolloverRange
        rolloverButton.setText(f'Unmark rollover range [{start}, {end}]')
        inBetweenButton.setEnabled(False)
        return
      else:
        rolloverButton.setText('Start marking rollover')
      inBetweenRange, currentRangeIdx = getRange(frame, slider.inBetweenRanges)
      if inBetweenRange is not None:
        start, end = inBetweenRange
        inBetweenButton.setText(f'Unmark in between range [{start}, {end}]')
        rolloverButton.setEnabled(False)
      else:
        inBetweenButton.setText('Start marking in between')

    def btnClicked(btn, data):
      currentFrame = slider.value()
      if currentRangeIdx is not None:
        del data[currentRangeIdx]
        slider.update()
      elif self._currentRange is None:
        btn.setText(btn.text().replace('Start', 'Stop'))
        otherBtn = rolloverButton if btn is not rolloverButton else inBetweenButton
        otherBtn.setEnabled(False)
        saveButton.setEnabled(False)
        self._currentRange = [currentFrame, currentFrame]
        data.append(self._currentRange)
      else:
        self._currentRange.sort()
        data.sort()
        mergedRanges = []
        for range_ in data:
          if not mergedRanges:
            mergedRanges.append(range_)
          else:
            if mergedRanges[-1][1] == range_[0] - 1:
              mergedRanges[-1][1] = range_[1]
            else:
              mergedRanges.append(range_)
        data[:] = mergedRanges
        slider.update()

        self._currentRange = None
        saveButton.setEnabled(self._originalRolloverClassification != self._rolloverClassification)
      updateRollover(currentFrame)

    rolloverButton = QPushButton('Start marking rollover')
    rolloverButton.clicked.connect(lambda: btnClicked(rolloverButton, slider.rolloverRanges))
    rolloverLayout.addWidget(rolloverButton, alignment=Qt.AlignmentFlag.AlignCenter)
    inBetweenButton = QPushButton('Start marking in between')
    inBetweenButton.clicked.connect(lambda: btnClicked(inBetweenButton, slider.inBetweenRanges))
    rolloverLayout.addWidget(inBetweenButton, alignment=Qt.AlignmentFlag.AlignCenter)

    def saveRolloverClassification():
      with open(self._rolloverClassificationPath, 'w') as f:
        json.dump(self._rolloverClassification, f, indent=2)
      saveButton.setEnabled(False)
      self._originalRolloverClassification = copy.deepcopy(self._rolloverClassification)

    self._saveButton = saveButton = QPushButton('Save changes')
    saveButton.setEnabled(False)
    saveButton.clicked.connect(saveRolloverClassification)
    rolloverLayout.addWidget(saveButton, alignment=Qt.AlignmentFlag.AlignCenter)
    rolloverWidget = QWidget()
    rolloverWidget.setLayout(rolloverLayout)
    layout.addWidget(rolloverWidget, alignment=Qt.AlignmentFlag.AlignCenter)

    startPageBtn = QPushButton("Go to the start page", self)
    startPageBtn.clicked.connect(lambda: self.setFolder(None) or controller.show_frame("StartPage"))
    layout.addWidget(startPageBtn, alignment=Qt.AlignmentFlag.AlignCenter)

    centralWidget = QWidget()
    centralWidget.setLayout(layout)
    self.addWidget(tree)
    self._centralWidget = wrapperWidget = QWidget()
    wrapperLayout = QHBoxLayout()
    wrapperLayout.addWidget(centralWidget, alignment=Qt.AlignmentFlag.AlignCenter)
    wrapperWidget.setLayout(wrapperLayout)
    wrapperWidget.showChildren = lambda: [child.show() for child in centralWidget.findChildren(QWidget) if child is not startPageBtn and child is not self._messageLabel]
    wrapperWidget.hideChildren = lambda: [child.hide() for child in centralWidget.findChildren(QWidget) if child is not startPageBtn]
    scrollArea = QScrollArea()
    scrollArea.setFrameShape(QFrame.Shape.NoFrame)
    scrollArea.setWidgetResizable(True)
    scrollArea.setWidget(wrapperWidget)
    self.addWidget(scrollArea)
    self.setStretchFactor(1, 1)
    self.setChildrenCollapsible(False)
    wrapperWidget.hideChildren()

  def _wellSelected(self, idx):
    if self._currentRange is not None:
      try:
        self._slider.rolloverRanges.remove(self._currentRange)
        self._slider.inBetweenRanges.remove(self._currentRange)
      except ValueError:
        pass
    self._currentRange = None
    self._slider.rolloverRanges = self._rolloverClassification[f'{idx + 1}']['rollover']
    self._slider.inBetweenRanges = self._rolloverClassification[f'{idx + 1}']['inBetween']
    self._slider.valueChanged.emit(self._slider.value())
    self._slider.update()

  def _findResultsFile(self, folder):
    reference = os.path.join(self.controller.ZZoutputLocation, os.path.join(folder, 'results_' + folder + '.txt'))
    if os.path.exists(reference):
      return reference
    mypath = os.path.join(self.controller.ZZoutputLocation, folder)
    if not os.path.exists(mypath):
      return None
    resultsFile = next((f for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath, f)) and f.startswith('results_')), None)
    if resultsFile is None:
      return None
    return os.path.join(self.controller.ZZoutputLocation, os.path.join(folder, resultsFile))

  def setFolder(self, name):
    self._currentRange = None
    self._title_label.setText(name)
    if name is None:
      self._tree.hide()
      filesystemModel = self._tree.model().sourceModel()
      filesystemModel.setRootPath(None)
      filesystemModel.setRootPath(self.controller.ZZoutputLocation)  # force refresh of the model
      self._tree.setRootIndex(self._tree.model().mapFromSource(filesystemModel.index(filesystemModel.rootPath())))
      self._centralWidget.hideChildren()
      self._tree.selectionModel().reset()
      self._tree.show()
      return
    else:
      frameInfo = getFramesCallback('', name, -1, 0, False, 0, ZZoutputLocation=self.controller.ZZoutputLocation)
      if frameInfo is None:
        self._centralWidget.hideChildren()
        self._messageLabel.show()
        return
      self._getFrame, frameRange, _, toggleTrackingPoints, wellPositions, wellShape, _, _ = frameInfo

      self._rolloverClassificationPath = os.path.join(self.controller.ZZoutputLocation, name, 'rolloverManualClassification.json')
      if os.path.exists(self._rolloverClassificationPath):
        with open(self._rolloverClassificationPath) as f:
          self._rolloverClassification = json.load(f)
        self._saveButton.setEnabled(False)
      else:
        self._rolloverClassification = {str(wellIdx + 1): {"rollover": [], "inBetween": []} for wellIdx in range(len(wellPositions))}
        self._saveButton.setEnabled(True)
      self._originalRolloverClassification = copy.deepcopy(self._rolloverClassification)

      self._centralWidget.hideChildren()

      self._slider.setRange(*frameRange)
      self._slider.setValue(self._slider.minimum())

      pixmap = QPixmap(util._cvToPixmap(self._getFrame(self._slider)))
      label = QLabel()
      label.setPixmap(pixmap)
      parentLayout = self._frame.parentWidget().layout()
      parentLayout.replaceWidget(self._frame, label)

      if wellShape is None:
        self._frame.wellShape = None
        self._frame.setWellPositions(None)
      else:
        self._frame.wellShape = wellShape
        self._frame.setWellPositions([(position['topLeftX'], position['topLeftY'], position['lengthX'], position['lengthY'])
                                      for idx, position in enumerate(wellPositions)])

      self._centralWidget.showChildren()

      size = pixmap.size().scaled(label.size(), Qt.AspectRatioMode.KeepAspectRatio)
      label.hide()
      self._frame.sizeHint = lambda: size
      self._frame.setMinimumSize(size)
      self._frame.setMaximumSize(size)
      self._frame.viewport().setFixedSize(size)
      self._frame.setOriginalPixmap(pixmap)
      parentLayout.replaceWidget(label, self._frame)
      label.setParent(None)
